;MACROS
;This macros collection helps to perform some usual tasks with XMM registers and
;instructions supported by SSE2.

;The macro had to be created because masm 6.15 does not recognize
;movsd instruction with XMM registers. It only takes it as movsd which is
;used with 32-bit registers to copy memory directly from src (ESI) to dest (ESI)
;SSE2 movsd is represented in bytes like this: F2 0F 10 /r.
;There is a similar problem with cmpsd. It is dealt with in xmm_cmpsd macro 
;The idea of the movsd macro is as follows.
;MASM 6.15 can properly encode movss xmm,xmm/m64.
;MOVSS is same as MOVSD except the first byte is F3 and not F2.
;So the macro makes MASM encode movss with our arguments,
;then it returns to the beginning, inserts F2 in there and goes to the end
;of the instructions being created.
;CREDITS:
;This macro was created by Christian Ludloff (ludloff@sandpile.org)
xmm_movsd macro ops:vararg
	local beg_instr
	local end_instr
	beg_instr: movss &ops
	end_instr: org beg_instr
	db 0F2h
	org end_instr
endm

xmm_cmpsd macro ops:vararg
	local beg_instr
	local end_instr
	beg_instr: cmpss &ops
	end_instr: org beg_instr
	db 0F2h
	org end_instr
endm

shuf macro x3:REQ,x2:REQ,x1:REQ,x0:REQ
	if (x3 GT 3 OR x2 GT 3 OR x1 GT 3 OR x0 GT 3)
		.ERR <Constants must be in the range of [0..3]>
	else
		exitm<(x3 shl 6) + (x2 shl 4) + (x1 shl 2) + x0>
	endif
endm

;SIMD magic constants
SIMD_COPY		equ shuf(3,2,1,0)	;0E4h	;11100100b
SIMD_XCHG		equ shuf(1,0,3,2)	;04Eh	;01001110b
SIMD_DXCHG		equ shuf(2,3,0,1)	;0B1h	;10110001b	;xchg dwords in lower QWORD and xchg dwords in upper QWORD 
SIMD_ROL		equ shuf(2,1,0,3)	;093h	;10010011b
SIMD_ROR		equ shuf(0,3,2,1)	;039h	;00111001b
SIMD_SWAP		equ shuf(0,1,2,3)	;01Bh	;00011011b
SIMD_SWAPLO		equ shuf(3,2,0,1)	;0E1h	;11100001b
SIMD_SWAPHI		equ shuf(2,3,1,0)	;0B4h	;10110100b
SIMD_COMBLO		equ shuf(1,0,1,0)	;044h	;01000100b	;with shufps combines lo from src and lo from dest
SIMD_COMBHI		equ shuf(3,2,3,2)	;0EEh	;11101110b	;with shufps combines hi from src and hi from dest
SIMD_FILL0		equ shuf(0,0,0,0)	;000h	;00000000b	;fill entire XMM with 0th (lowest) dword
SIMD_FILL1		equ shuf(1,1,1,1)	;055h	;01010101b	;fill entire XMM with 1st dword
SIMD_FILL2		equ shuf(2,2,2,2)	;0AAh	;10101010b	;fill entire XMM with 2nd dword
SIMD_FILL3		equ shuf(3,3,3,3)	;0FFh	;11111111b	;fill entire XMM with 3rd dword

;SSE,SSE2 macros

;movdqa seems to run faster.
xmm_mov	macro regDest:REQ,regSrc:REQ
	movdqa	regDest,regSrc
endm

;this one seemed to work slower on P4 2.4GHz in spite of Intel Latencies Info
xmm_move macro regDest:REQ,regSrc:REQ
	pshuflw	regDest,regSrc,SIMD_COPY
endm

;exchange low and high qwords from src register and copy the resutl into dest register
xmm_xchg macro regDest:REQ,regSrc:REQ
	pshufd	regDest,regSrc,SIMD_XCHG
endm

;exchange dwords in both high and low qwords
xmm_dxchg macro regDest:REQ,regSrc:REQ
	pshufd	regDest,regSrc,SIMD_DXCHG
endm

;exchange hi and lo dwords of lo qword in regSrc and copy the result into regDest
;hi qword is left intact
xmm_xchgLq macro regDest:REQ,regSrc:REQ
	pshuflw	regDest,regSrc,SIMD_XCHG
endm

;exchange hi and lo dwords of hi qword in regSrc and copy all oword into regDest
;lo qword is left intact
xmm_xchgHq macro regDest:REQ,regSrc:REQ
	pshufhw	regDest,regSrc,SIMD_XCHG
endm

;take dwords from src register, reverse their order (3,2,1,0>0,1,2,3) and save to dest reg
xmm_swapdw	macro regDest:REQ,regSrc:REQ
	pshufd	regDest,regSrc,SIMD_SWAP
endm

;rol XMM,32 rotate by dwords to the left, 32 bits at a time
xmm_rol macro regDest:REQ,regSrc:REQ
	pshufd	regDest,regSrc,SIMD_ROL
endm

;rol XMM.lo,16 rotate low QWORD of XMM by words to the left, 16 bits at a time
xmm_rolLq macro regDest:REQ,regSrc:REQ
	pshuflw	regDest,regSrc,SIMD_ROL
endm

;rol XMM.hi,16 rotate high QWORD of XMM by words to the left, 16 bits at a time
xmm_rolHq macro regDest:REQ,regSrc:REQ
	pshufhw	regDest,regSrc,SIMD_ROL
endm

;ror XMM,32 rotate by dwords to the right, 32 bits at a time
xmm_ror macro regDest:REQ,regSrc:REQ
	pshufd	regDest,regSrc,SIMD_ROR
endm

;ror XMM.lo,16 rotate low QWORD of XMM by words to the right, 16 bits at a time
xmm_rorLq macro regDest:REQ,regSrc:REQ
	pshuflw	regDest,regSrc,SIMD_ROR
endm

;ror XMM.hi,16 rotate high QWORD of XMM by words to the right, 16 bits at a time
xmm_rorHq macro regDest:REQ,regSrc:REQ
	pshufhw	regDest,regSrc,SIMD_ROR
endm

;fill the rest of the register with contents of one of the dwords
;fillpos is the number of the dword in range 0 through 3
xmm_filldw macro reg:REQ,fillpos:REQ
	LOCAL	inst
	if fillpos GT 3
		.ERR <Too big fill position of dword>
		exitm
	endif
	if fillpos EQ 1
		fillpos=SIMD_FILL1
		goto inst
	endif
	if fillpos EQ 2
		fillpos=SIMD_FILL2
		goto inst
	endif
	if fillpos EQ 3
		fillpos=SIMD_FILL3
	endif
:inst
	pshufd	reg,reg,fillpos
endm

;for xmm_nop you can specify the register that can be used inside
;if the register is not specified then XMM0 is used
xmm_nop macro reg
	ifdef reg
		;pshuflw	reg,reg,SIMD_COPY
		movdqa	reg,reg
	else
		;pshuflw	xmm0,xmm0,SIMD_COPY
		movdqa	xmm0,xmm0
	endif
endm

;set all bits in register to 0
xmm_zero macro reg:REQ
	pxor	reg,reg
endm

;set all bits in register to 1
xmm_minus1 macro reg:REQ
    pcmpeqd	reg,reg
endm

;take src register, invert its bits and save the result into dest reg
xmm_not macro regdest:REQ,regsrc:REQ
	if regdest EQ regsrc
		.err <source and destination registers cannot be same>
	else
		xmm_minus1 regdest
		pxor	regdest,regsrc
	endif
endm

;does not preserve hi qword, but sets hi qword to 0
;it is fast however
xmm_lo2loX	macro regDest:REQ,regSrc:REQ
	movq	regDest,regDest
endm

;preserves hi qword if dest is register; ATTENTION: zeroes our hi qword if regSrc is memory
xmm_lo2lo	macro regDest:REQ,regSrc:REQ
	xmm_movsd	regDest,regSrc
endm

xmm_hi2hi	macro regDest:REQ,regSrc:REQ
	shufps	regDest,regSrc,SIMD_COPY
endm

;this one moves src hi qword into dest hi qword but dest hi qword is shifted right into lo qword
xmm_hi2hishr macro regDest:REQ,regSrc:REQ
	punpckhqdq	regDest,regSrc
endm

;All of these instructions do the same. But the last one seems to work faster 
xmm_lo2hi	macro regDest:REQ,regSrc:REQ
	;movlhps	regDest,regSrc
	;unpcklpd	regDest,regSrc	;these are basicly same
	punpcklqdq	regDest,regSrc	;these are basicly same
endm

;this one does not perserve hi part of destination register, but it is fast, uses MMX_SHFT
xmm_hi2loX	macro regDest:REQ,regSrc:REQ
	pshufd	regDest,regSrc,SIMD_XCHG
endm

;this one preserves hi qw in destination register intact
;it's one slow instruction, 6 tacts + 2 throughput, uses MMX_SHFT execution unit
xmm_hi2lo	macro regDest:REQ,regSrc:REQ
	movhlps	regDest,regSrc
endm

;copy lo 32-bits from src to dest
;all the other dwords in dest are left unchanged if src is reg; ATTENTION: zeroes out all higher dwords if regSrc is memory
xmm_mov32	macro regDest:REQ,regSrc:REQ
	movss	regDest,regSrc
endm

;these instructions set xmm6 and xmm7 to the const that I will use in most moves
xmm_setconst macro
	xmm_minus1	xmm7
	psllq	xmm7,3*8
	xmm_minus1	xmm6
	psrldq	xmm7,2
	pxor	xmm6,xmm7	
endm



;MMX macros (supported starting SSE,SSE2)

;the following 2 macros can be used with movq,movaps for parallel execution
mmx_mov	macro dest:REQ,src:REQ
	pshufw	dest,src,SIMD_COPY
endm

mmx_xchg macro regDest:REQ,regSrc:REQ
	pshufw		regDest,regSrc,SIMD_XCHG
endm

;this one does not perserve hi part of destination register, but it is faster
mmx_hi2loX	macro regDest:REQ,regSrc:REQ
	pshufw	regDest,regSrc,SIMD_XCHG
endm

;this one preserves hi dw in destination register intact, but it is slower than mmx_hi2lo
mmx_hi2lo	macro regDest:REQ,regSrc:REQ
	punpckhdq	regDest,regSrc
	mmx_xchgHiLo	regDest,regDest
endm

mmx_rol macro regDest:REQ,regSrc:REQ
	pshufw	regDest,regSrc,SIMD_ROL
endm

mmx_ror macro regDest:REQ,regSrc:REQ
	pshufw	regDest,regSrc,SIMD_ROR
endm

mmx_swapw	macro reg:REQ
	pshufw	reg,reg,SIMD_SWAP
endm

mmx_fillw macro reg:REQ,fillpos:REQ
	LOCAL	inst
	if fillpos GT 3
		.ERR Too big fill position of dword
		exitm
	endif
	if fillpos EQ 1
		fillpos=SIMD_FILL1
		goto inst
	endif
	if fillpos EQ 2
		fillpos=SIMD_FILL2
		goto inst
	endif
	if fillpos EQ 3
		fillpos=SIMD_FILL3
	endif
:inst
	pshufw	reg,reg,fillpos
endm
